from .display import \
    JsonLInteractiveClassifier, \
    TweetInteractiveClassifier, \
    prepare_google_credentials
from .analysis import \
    TweetAnalyzer,TweetMedia, TweetPhoto, TweetVideo
from .session import TSess
from .cache import Cache




import os, logging
import os.path
import datetime
import zlib
from enum import Enum
from hashlib import md5, sha1
from datetime import datetime
from sys import getsizeof


def get_size(obj: object, seen=None):
    """Recursively finds size of objects"""
    size = getsizeof(obj)
    if seen is None:
        seen = set()
    obj_id = id(obj)
    if obj_id in seen:
        return 0
    # Important mark as seen *before* entering recursion to gracefully handle
    # self-referential objects
    seen.add(obj_id)
    if isinstance(obj, dict):
        size += sum([get_size(v, seen) for v in obj.values()])
        size += sum([get_size(k, seen) for k in obj.keys()])
    elif hasattr(obj, '__dict__'):
        size += get_size(obj.__dict__, seen)
    elif hasattr(obj, '__iter__') and not isinstance(obj, (str, bytes, bytearray)):
        size += sum([get_size(i, seen) for i in obj])
    return size

    def request(self, key: K) -> V:
        pass

    def _get(self, key: K) -> V:
        value = self.request(key)
        store = Retrievable(value)
        self.CACHE.update({key: store})
        return value


class HashType(Enum):
    md5 = 1
    sha1 = 2


class Request():
    def __init__(
        self, uri: str, method: str = "GET",
        params: dict = {}, headers: dict = {}
    ):
        if type(uri) is not str:
            try:
                uri = str(uri)
            except:
                raise
        assert type(uri) is str, "uri most be a string!"
        assert type(method) is str, "method must be a string!"
        assert type(headers) is dict, "headers must be a dictionary!"
        self.URI = uri
        self.method = method.upper()
        self.params = self.order_dict(params)
        self.headers = self.order_dict(headers)

    @staticmethod
    def order_dict(dictionary: dict):
        keys = list(dictionary.keys())
        keys.sort()
        new_headers = {}
        for key in keys:
            new_headers.update({key: dictionary[key]})
        return new_headers


class Cache():
    TWO_YEARS_IN_DAYS = 780.50

    def __init__(
            self, cache_dir="./.tweet_cache/", hash_function=HashType.md5,
            soft_reload=False, refresh_rate=TWO_YEARS_IN_DAYS,
            compression_level: int = 3, hash_split=True, split_size=2, alt_hash=None
    ):
        assert type(hash_function) is HashType or (
            hash_function is None and alt_hash is not None), f"hash_function must be of type '{type(HashType.md5)}''"
        self.HASH_SPLIT = hash_split
        self.hash: HashType = hash_function
        self.SOFT_RELOAD: bool = soft_reload
        self.REFRESH_RATE: float = refresh_rate * 24 * \
            60 * 60  # Change from years to seconds
        self.ALT_HASH = alt_hash
        self.SPLIT_SIZE = split_size
        try:
            assert 0 <= compression_level <= 9, f"Compression level {compression_level} invalid! Defaulting to CmpLvl=3."
        except:
            compression_level = 0
        finally:
            self.COMPRESS_LEVEL = compression_level
        if not os.path.isdir(cache_dir):
            assert not os.path.isfile(
                cache_dir), f"'{cache_dir}' is a file not a directory!"
            os.makedirs(cache_dir)
        self.CACHE_DIR = cache_dir

    def uri_hash(self, uri: str, method: str = "GET", params: dict = {}, headers: dict = {}) -> str:
        tweet_request = Request(uri, method, params, headers)
        return self.request_hash(tweet_request)

    def request_hash(self, tweet_request: Request) -> str:
        def hash(x): return x
        if self.hash is HashType.md5:
            hash = md5
        elif self.hash is HashType.sha1:
            hash = sha1
        elif self.hash is None:
            return self.ALT_HASH(tweet_request)

        # if self.hash is HashType.md5:
        #     if header_hash is None:
        #         header_hash = "{" + md5(str(tweet_request.headers).encode('utf-8')).hexdigest() + "+"
        #     pre_hash_str = (tweet_request.method+"_"+tweet_request.URI + str(tweet_request.headers))
        #     return md5(pre_hash_str.encode('utf-8')).hexdigest()
        # elif self.hash is HashType.sha1:
        #     pre_hash_str = (tweet_request.method+"_"+tweet_request.URI + str(tweet_request.headers))
        #     return sha1(pre_hash_str.encode('utf-8')).hexdigest()

        # Hash header to avoid displaying public information
        if tweet_request.headers == {}:
            header_hash = "{}"
        else:
            header_hash = "{" + \
                hash(str(tweet_request.headers).encode(
                    'utf-8')).hexdigest() + "}"
        pre_hash_str = tweet_request.method + "_" + \
            tweet_request.URI + str(tweet_request.params) + header_hash
        return hash(pre_hash_str.encode('utf-8')).hexdigest()

    def request_stamp(self, tweet_request: Request) -> float:
        """
        Returns the milisecond presicion of the last time a record was cached.
        The integer part of the float is equivalent to seconds.

        Args:
            request (Request): request that includes URI and method

        Returns:
            float: Timestamp of the cache file.
        """
        if self.present(tweet_request=tweet_request):
            file_name = self.request_filename(tweet_request=tweet_request)
            return os.path.getmtime(file_name)
        return 0

    def needs_update(self, tweet_request: Request, refresh_rate: float = None) -> bool:
        """
        Determines if a value needs to be updated based on a refresh_rate 
        float number given in days based on the cach timestamp.

        Args:
            request (Request): Request object that references the resource.
            refresh_rate (float, optional): 
                Custom refresh_rate. Defaults to None to use self.REFRESH_RATE.

        Returns:
            bool: True if the resource need updating.
        """
        if self.SOFT_RELOAD:
            if refresh_rate is None:
                refresh_rate = self.REFRESH_RATE
            filename = self.request_filename(tweet_request=tweet_request)
            if os.path.isfile(filename):
                stamp = os.path.getmtime(filename=filename)
                now = datetime.now().timestamp()
                if now - stamp > refresh_rate:
                    return True
            else:
                return True
        return False

    def uri_filename(self, uri: str, method: str = "GET", params: dict = {}, headers: dict = {}):
        tweet_request = Cache.request_from_uri(uri, method, params, headers)
        return self.request_filename(tweet_request=tweet_request)

    def request_filename(self, tweet_request: Request) -> str:
        if not self.HASH_SPLIT:
            return os.path.join(self.CACHE_DIR, self.request_hash(tweet_request))
        else:
            hash_str = self.request_hash(tweet_request=tweet_request)
            n = self.SPLIT_SIZE
            relative_path = os.path.join(
                self.CACHE_DIR, *[hash_str[i:i+n] for i in range(0, len(hash_str), n)])
            os.makedirs(os.path.dirname(relative_path), exist_ok=True)
            return relative_path

    @staticmethod
    def request_from_uri(uri: str, method: str = "GET", params: dict = {}, headers: dict = {}) -> Request:
        tweet_request = Request(uri, method, params, headers)
        return tweet_request

    def present(self, tweet_request: Request) -> bool:
        file_name = self.request_filename(tweet_request=tweet_request)
        return os.path.isfile(file_name)

    def check(self, uri: str, method: str = "GET", params: dict = {}, headers: dict = {}):
        tweet_request = Cache.request_from_uri(uri, method, params, headers)
        return self.available(tweet_request)

    def available(self, tweet_request: Request):
        try:
            assert type(
                tweet_request) is Request, f"Invalid type {type(tweet_request)} as `request` argument for `available` method, need a Request object."
        except:
            if type(tweet_request) is str:
                tweet_request = Request(tweet_request)
            else:
                raise
        return self.present(tweet_request) and not self.needs_update(tweet_request)

    def get(self, uri: str, method: str = "GET", params: dict = {}, headers: dict = {}) -> str:
        tweet_request = Request(uri, method, params, headers)
        if self.available(tweet_request):
            with open(self.request_filename(tweet_request), 'rb') as cache:
                data = cache.read()
            dc_data = zlib.decompress(data)
            return dc_data.decode("utf-8")
        else:
            return "Unavailable"

    def store_bytes(self, uri: str, value: bytes, method: str = "GET", params: dict = {}, headers: dict = {}):
        tweet_request = Request(uri, method, params, headers)
        data = zlib.compress(value, level=self.COMPRESS_LEVEL)
        with open(self.request_filename(tweet_request), "wb") as cache:
            cache.write(data)
        logging.debug(f"Stored at: {self.request_filename(tweet_request)}")

    def store(self, uri: str,  value: str, method: str = "GET", params: dict = {}, headers: dict = {}):
        logging.debug("Storing...")
        self.store_bytes(uri, value.encode("utf-8"), method, params, headers)







import json, logging
from time import sleep
import requests
from .cache import Cache, Request
from typing import Union, Tuple, List


# Source https://developer.twitter.com/en/docs/twitter-api/tweets/lookup/api-reference/get-tweets-id
TWEET_BY_ID_URL = "https://api.twitter.com/2/tweets/"


class BearerAuth(requests.auth.AuthBase):
    """ ReadOnly Bearer Token authentication. 
    It only requires the bearer token to work.
    """

    def __init__(self, token: str):
        self.token = token

    def __call__(self, r: Request):
        r.headers["authorization"] = "Bearer " + self.token
        return r


class TSess():
    BASE_URL_11 = "https://api.twitter.com/1.1/statuses/lookup.json"

    def __init__(
        self,
        bearer_token: str,
        expansions: List[str] = [
            "attachments.media_keys",
            "author_id",
            "entities.mentions.username",
            "referenced_tweets.id",
            "referenced_tweets.id.author_id",
        ],
        media_fields: List[str] = [
            "type",
            "duration_ms",
            "preview_image_url",
            "public_metrics",

        ],
        place_fields: List[str] = [
            "country",
            "country_code",
            "full_name",
            "geo",
            "id",
            "name",
            "place_type",
            "contained_within",
        ],
        user_fields: List[str] = [
            "username",
            "name",
            "id",
            "created_at",
            "description",
            "pinned_tweet_id",
        ],
        tweet_fields: List[str] = [
            "lang",
            "author_id",
            "id",
            "public_metrics",
            "possibly_sensitive",
            "conversation_id",
            "in_reply_to_user_id",
            "referenced_tweets",
            "source",
            "text",
            "geo",
        ],
        cache_dir: str = "./.tweet_bearer_cache/",
        error_log=".tweet_error.json",
        compression_level: int = 3,
        sleep_time=1.0,
        hash_split=False,
    ):
        self.auth = BearerAuth(bearer_token)
        self.cache = Cache(cache_dir=cache_dir, soft_reload=False,
                           compression_level=compression_level, hash_split=hash_split)
        self.ERROR_LOG = error_log
        self.SLEEP_TIME = sleep_time
        try:
            handler = open(self.ERROR_LOG, 'r')
            self.ERROR_DICT: dict = json.load(handler)
            handler.close()
        except:
            self.ERROR_DICT = {}

        self.PARAMS = {
            'expansions': ','.join(expansions),
            'media.fields': ','.join(media_fields),
            'tweet.fields': ','.join(tweet_fields),
            'place.fields': ','.join(place_fields),
            'user.fields': ','.join(user_fields),
        }
        self.failed_ids = []

    def load_tweet_batch(
        self, ids: List[str],
        base_url="https://api.twitter.com/2/tweets"
    ) -> Union[str, None]:
        # https://api.twitter.com/2/tweets?ids=1228393702244134912,1227640996038684673,1199786642791452673&tweet.fields=created_at&expansions=author_id&user.fields=created_at
        params = self.PARAMS.copy()
        params.update({"ids": ','.join(ids)})
        # URI = self.generate_URI(base_url, params)
        if self.cache.check(base_url, params=params):
            return self.cache.get(base_url, params=params)
        else:
            response = requests.get(base_url, params=params, auth=self.auth)
            if response.status_code == 200:
                self.cache.store(base_url, response.text, params=params)
                sleep(self.SLEEP_TIME)
                return response.text
        return None

    @staticmethod
    def generate_URI(base_url: str, params: dict, v2: bool = True):
        params = params.copy()
        params = Request.order_dict(params)
        uri = base_url
        if len(params.keys()) > 0:
            uri += "?"
            for key in params.keys():
                if hasattr(params[key], '__iter__') and type(params[key]) is not str:
                    uri += key + "=" + ','.join(params[key]) + "&"
                else:
                    uri += key + "=" + params[key] + "&"
        return uri[:-1]

    @staticmethod
    def generate_URI_11(tweet_id: str, params: dict = {}):
        new_params = {
            "id": tweet_id,
            "include_entities": True,
            "tweet_mode": "extended",  # Testing to include Extended Entities
            "trim_user": False,
        }
        params.update(new_params)
        return TSess.BASE_URL_11, params

    @staticmethod
    def generate_batch_URI_11(tweet_ids: List[str], params: dict = {}):
        tweet_ids_str = ",".join(tweet_ids)
        new_params = {
            "id": tweet_ids_str,
            "include_entities": True,
            "tweet_mode": "extended",  # Testing to include Extended Entities
            "trim_user": False,
        }
        params.update(new_params)
        return TSess.BASE_URL_11, params

    def load_request(
        self, base_url: str, params: dict,
        is_tweet: bool = True, is_v2: bool = True
    ) -> Tuple[str, int]:
        if self.cache.check(base_url, params=params):
            logging.debug("Value in Cache")
        elif id in self.ERROR_DICT.keys():
            logging.debug("Previous Error Found!")
            return json.dumps(self.ERROR_DICT[id][2]), self.ERROR_DICT[id][1]
        else:
            logging.debug("Need to request value")
            sleep(self.SLEEP_TIME)
            response = requests.get(base_url, params=params, auth=self.auth)
            response.status_code
            if response.status_code == 200:
                data = response.text
                sleep(0.0005)
                r: List[dict] = json.loads(data)
                if is_tweet and is_v2:
                    if type(r) is list:
                        assert len(r) > 0, "List is empty! Response was empty list."
                        r = r[0]
                sleep(0.0005)
                if "errors" in r.keys() and "data" not in r.keys() and is_tweet:
                    error: str = r["errors"][0]["title"]
                    logging.debug(f"{id} - Twitter Error Returned: {error}")
                    hash = self.cache.uri_hash(base_url, params=params)
                    error_code = 440
                    self.ERROR_DICT.update({id: (hash, error_code, r)})
                    with open(self.ERROR_LOG, "w") as handler:
                        json.dump(self.ERROR_DICT, handler, indent=2)
                    return data, error_code  # Using code 440 for any Twitter API error code found
                else:
                    self.cache.store(uri=base_url, value=data,
                                     params=params, method="GET")
            else:
                hash = self.cache.uri_hash(base_url, params=params)
                error_code = response.status_code
                self.ERROR_DICT.update({id: (hash, error_code, response.text)})
                with open(self.ERROR_LOG, "w") as handler:
                    json.dump(self.ERROR_DICT, handler, indent=2)
                logging.debug(f"Could not load tweet: {response.reason}")
                return response.text, response.status_code
        return self.cache.get(base_url, params=params), 200

    def load_tweet_11(self, id: str, v2: bool = True) -> Tuple[str, int]:
        if type(id) is int:
            id = str(id)
        elif type(id) is str:
            assert id.isnumeric(
            ), f"id {type(id)} provided is not a valid string representation of an integer."
        else:
            raise Exception(
                f"Invalid id type: {type(id)}. Only <class int> and <class str> are acceptable.")
        if v2:
            base_url, params = TSess.generate_URI_11(id, params=self.PARAMS)
        else:
            base_url, params = TSess.generate_URI_11(id, params={})

        return self.load_request(base_url=base_url, params=params)

    def load_tweet_batch_11(self, ids: List[str], v2: bool = True) -> Tuple[str, int]:
        for id in ids:
            try:
                assert id.isnumeric(
                ), f"id {type(id)} provided is not a valid string representation of an integer."
            except:
                raise Exception(
                    f"Invalid id type: {type(id)}. Only numeric <class str> are acceptable.")
        if v2:
            params = self.PARAMS
        else:
            params = {}
        base_url, params = TSess.generate_batch_URI_11(ids, params=params)

        return self.load_request(base_url=base_url, params=params)

    def load_tweet(self, id: str, base_url=TWEET_BY_ID_URL) -> Tuple[str, int]:
        if type(id) is int:
            id = str(id)
        elif type(id) is str:
            assert id.isnumeric(
            ), f"id {type(id)} provided is not a valid string representation of an integer."
        else:
            raise Exception(
                f"Invalid id type: {type(id)}. Only <class int> and <class str> are acceptable.")
        url = base_url + id
        headers = {}
        logging.debug(f"Requesting tweet {id} with uri: '{url}'")
        return self.load_request(base_url=url, params=self.PARAMS)
